/*
Bibliothèque de fonctions pour calculer des Transformées de Fourier

Copyright (C) 2012 Observatoire de Paris-Meudon

Ce programme est un logiciel libre ; vous pouvez le redistribuer et/ou le modifier conformément aux dispositions de la Licence Publique Générale GNU, telle que publiée par la Free Software Foundation ; version 2 de la licence, ou encore (à votre choix) toute version ultérieure.

Ce programme est distribué dans l'espoir qu'il sera utile, mais SANS AUCUNE GARANTIE ; sans même la garantie implicite de COMMERCIALISATION ou D'ADAPTATION A UN OBJET PARTICULIER. Pour plus de détail, voir la Licence Publique Générale GNU.

Vous devez avoir reçu un exemplaire de la Licence Publique Générale GNU en même temps que ce programme ; si ce n'est pas le cas, écrivez à la Free Software Foundation Inc., 675 Mass Ave, Cambridge, MA 02139, Etats-Unis.
*/

/**
 * Bibliothèque de fonctions pour calculer des Transformées de Fourier
 * fonctions publiques : TFLib.fft_rec et TFLib.transfo2D_FFT
 */

function TFLib(f_r, f_i) {
    this.tw_cos = null; // double[][]
    this.tw_sin = null; // double[][]
    this.f_r = f_r; // double[]
    this.f_i = f_i; // double[]
}

//protected void bit_reverse() {
TFLib.prototype.bit_reverse = function() {
    this.bit_reverse2(this.f_r);
    this.bit_reverse2(this.f_i);
}

//protected static void bit_reverse(double[] t) {
TFLib.prototype.bit_reverse2 = function(t) {
    var j, temp, m, i;
    j = 0;
    for (i=0; i<t.length-1; i++) {
        if (j > i) {
            temp = t[j];
            t[j] = t[i];
            t[i] = temp;
        }
        m = Math.floor(t.length / 2);
        while (j >= m) {
            j -= m;
            m = Math.floor(m / 2);
        }
        j += m;
    }
}

TFLib.prototype.precalcul_tw = function(t) {
    var n, log2n, i, tw, m, log2m;
    n = this.f_r.length;
    log2n = Math.floor(Math.log(n)/Math.log(2));
    //tw_cos = new double[n/2][log2n+1];
    tw_cos = new Array(Math.floor(n/2));
    for (i=0; i<n/2; i++)
        tw_cos[i] = new Array(log2n+1);
    //tw_sin = new double[n/2][log2n+1];
    tw_sin = new Array(Math.floor(n/2));
    for (i=0; i<n/2; i++)
        tw_sin[i] = new Array(log2n+1);
    m = n;
    log2m = log2n;
    while (m > 1) {
        for (i=0; i<m/2; i++) {
            tw = 2*Math.PI*i/m;
            tw_cos[i][log2m] = Math.cos(tw);
            tw_sin[i][log2m] = Math.sin(tw);
        }
        m = Math.floor(m / 2);
        log2m--;
    }
}

// avec precalcul_tw
//protected void dif(boolean inverse, int ind, int log2n) {
TFLib.prototype.dif = function(inverse, ind, log2n) {
    var i, np, ind2, e_r, e_i, o_r, o_i, tw_r, tw_i, tmp_r, tmp_i;
    
    np = 1 << (log2n - 1);
    ind2 = ind + np;
    
    for (i=0; i<np; i++) {
        // e = f[ind+i] + f[ind2+i];
        e_r = this.f_r[ind+i] + this.f_r[ind2+i];
        e_i = this.f_i[ind+i] + this.f_i[ind2+i];
        
        tw_r = tw_cos[i][log2n];
        tw_i = -tw_sin[i][log2n];
        if (inverse)
            tw_i = -tw_i;
        
        tmp_r = this.f_r[ind+i] - this.f_r[ind2+i];
        tmp_i = this.f_i[ind+i] - this.f_i[ind2+i];
        
        // o = (f[ind+i] - f[ind2+i]) * e^(-2*j*pi*i/n);
        o_r = tmp_r * tw_r - tmp_i * tw_i;
        o_i = tmp_r * tw_i + tmp_i * tw_r;
        
        // f[ind+i] = e;
        this.f_r[ind+i] = e_r;
        this.f_i[ind+i] = e_i;
        
        // f[ind2+i] = o;
        this.f_r[ind2+i] = o_r;
        this.f_i[ind2+i] = o_i;
    }
    
    if (log2n > 1) {
        this.dif(inverse, ind, log2n-1);
        this.dif(inverse, ind2, log2n-1);
    }
}


/**
 * FFT récursive un peu optimisée.
 * la taille de f (f_r et f_i) doit être une puissance de 2.
 * avec precalcul_tw.
 */
//public static void fft_rec(boolean inverse, double[] f_r, double[] f_i) {
TFLib.fft_rec = function(inverse, f_r, f_i) {
    var lib, log2n;
    
    lib = new TFLib(f_r, f_i);
    lib.precalcul_tw();
    log2n = Math.floor(Math.log(f_r.length)/Math.log(2));
    lib.dif(inverse, 0, log2n);
    lib.bit_reverse();
}


/**
 * FFT en 2D
 */
//public static void transfo2D_FFT(double[][] imagep, double[][] figured, boolean log, boolean quadrants) {
TFLib.transfo2D_FFT = function(imagep, figured, log, quadrants) {
    var i, j, dimx, dimy, fk, kx, f_r, f_i, v, i2, ky2;
    
    dimx = imagep[0].length;
    dimy = imagep.length;
    for (i=0; i<dimx; i++)
        for (j=0; j<dimy; j++)
            figured[i][j] = 0;
    //double[][][] fk = new double[dimx][dimy][2];
    fk = new Array(dimx);
    for (i=0; i<dimx; i++) {
        fk[i] = new Array(dimy);
        for (j=0; j<dimy; j++)
            fk[i][j] = new Array(2);
    }
    for (kx=0; kx<dimx; kx++) {
        f_r = new Array(dimy);
        f_i = new Array(dimy);
        for (i=0; i<dimy; i++) {
            f_r[i] = imagep[kx][i];
            f_i[i] = 0;
        }
        
        TFLib.fft_rec(false, f_r, f_i);
        
        for (i=0; i<dimy; i++) {
            fk[kx][i][0] = f_r[i];
            fk[kx][i][1] = f_i[i];
        }
    }
    for (ky=0; ky<dimy; ky++) {
        f_r = new Array(dimx);
        f_i = new Array(dimx);
        for (i=0; i<dimx; i++) {
            f_r[i] = fk[i][ky][0];
            f_i[i] = fk[i][ky][1];
        }
        
        TFLib.fft_rec(false, f_r, f_i);
        
        for (i=0; i<dimx; i++) {
            if (log)
                v = Math.log(Math.sqrt(f_r[i]*f_r[i] + f_i[i]*f_i[i]));
            else
                v = Math.sqrt(f_r[i]*f_r[i] + f_i[i]*f_i[i]);
            if (quadrants) {
                // réorganisation des coordonnées
                if (i >= Math.floor(dimx/2))
                    i2 = i - Math.floor(dimx/2);
                else
                    i2 = i + Math.floor(dimx/2);
                if (ky >= Math.floor(dimy/2))
                    ky2 = ky - Math.floor(dimy/2);
                else
                    ky2 = ky + Math.floor(dimy/2);
                figured[i2][ky2] = v;
            } else
                figured[i][ky] = v;
        }
    }
}

